# MoabDB

class MoabError(Exception):
    """Base class for exceptions in this module."""
    pass


class MoabVersionError(MoabError):
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class MoabRequestError(MoabError):
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class MoabInternalError(MoabError):
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class MoabResponseError(MoabError):
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class MoabHttpError(MoabError):
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message



# MoabDB

_api_key = ""
_api_username = ""
_dx_url = "https://api.moabdb.com/"


def login(username: str, key: str):
    """
    Logs in to the API to provide data that is not publically available
    :param key: The API key to use
    :return: None
    """

    global _api_key
    global _api_username

    _api_key = key
    _api_username = username


def set_dx_url(url: str):
    """
    Sets the URL for the MoabDB API
    Only useful for testing
    :param url: The URL to use
    :return: None
    """

    global _dx_url
    _dx_url = url



# Jackson Coxson

from moabdb import protocol_pb2, send_request, check_version
import pandas as pd
import io

print(check_version())

for i in range(1, 100):
    req = protocol_pb2.Request()
    req.symbol = "AAPL"
    req.start = 1543554489
    req.end = 1548825309
    req.datatype = "intraday_stocks"

    res = send_request(req)

    if res.code == 200:
        # Slap the data into a pandas dataframe
        pq_file = io.BytesIO(res.data)
        df = pd.read_parquet(pq_file)
        print(df)
    else:
        print(res.code)
        print(res.message)



# MoabDB

import requests
import io
from . import globals
from . import __version__
from . import protocol_pb2
from . import errors
from .timewindows import _get_unix_dates


from base64 import b64encode, b64decode
import pandas as pd


def hello():
    print("Welcome to Moab, where all data is exchanged!!")


def _check_version() -> bool:
    """
    Checks the server's version, compairing the current version
    """
    try:
        res = requests.get(globals._dx_url + 'client_version/')
        return (res.text == __version__)
    except:
        raise errors.MoabHttpError("Unable to connect to server")


def _check_access() -> bool:
    """
    Checks if the user has set the global credentials.
    This function is just a local sanity check, the server will
    check the credentials again.
    """
    if globals._api_key == "":
        return False
    elif globals._api_username == "":
        return False


def _send_request(request: protocol_pb2.Request) -> protocol_pb2.Response:
    """
    Sends a request to the MoabDB API
    :param Request: The request to send
    :return: The response from the server
    """
    s = request.SerializeToString()
    headers = {
        'x-req': b64encode(s)
    }

    try:
        res = requests.get(globals._dx_url + 'request/v1/', headers=headers)
        if res.status_code != 200:
            raise errors.MoabRequestError(
                "Server returned error code: " + str(res.status_code))
        res = protocol_pb2.Response().FromString(b64decode(res.text))
        return res
    except:
        raise errors.MoabHttpError("Unable to connect to server")

def _server_req(ticker, start, end, datatype):
    # Request data from moabdb server
    req = protocol_pb2.Request()
    req.symbol = ticker
    req.start = start
    req.end = end
    req.datatype = datatype

    res = _send_request(req)

    if res.code == 200:
        # Place data into a dataframe
        pq_file = io.BytesIO(res.data)
        try:
            df = pd.read_parquet(pq_file)
            return (df)
        except:
            raise errors.MoabResponseError("Server returned invalid data")
    else:
        raise errors.MoabRequestError(
            "Server returned error code: " + str(res.code))


def get_equity(tickers, sample="1m",
               start=None, end=None,
               intraday=False):
    """
    get_equity paramaters:
        tickers: str OR list of strings
              Ex: "NVDA" or ["NVDA","AMD"]
        sample: str
            Sample length, required if "start" or "end" is missing
            Enter as number then frequency string (D, W, M, Y)
              Ex: "30d", "3m", "5y", etc.
        start: str
            Beginning date of sample, required if "end" or "sample" is missing
              Ex: '2020-01-01'
        end: str
            Ending date of sample, required if "start" or "sample" is missing
              Ex: '2022-05-01''
        intraday: bool
            True to return intraday data
            Default is 'False' to return end-of-day data
            See moabdb.com to look at subscriptions for intraday access
        api_key: str
            api_key required for intraday data
    """

    # Check intraday authorization
    if (intraday == True):
        is_authorized = _check_access()
        equity_freq = "intraday_stocks"
        if is_authorized == 0:
            # TODO: Add error code
            print("Error: Subscription required to access intraday data")
            intraday = False
            equity_freq = "daily_stocks"
    else:
        equity_freq = "daily_stocks"

    # String time to integer time
    start_tm, end_tm = _get_unix_dates(sample, start, end)

    # Single ticker request
    if isinstance(tickers, str):
        return_db = _server_req(
            str.upper(tickers), start_tm, end_tm, equity_freq)

    # List of tickers request
    elif isinstance(tickers, list):
        compile_tickers = []
        for tic in tickers:
            compile_tickers.append(_server_req(
                str.upper(tic), start_tm, end_tm, equity_freq))
        return_db = pd.concat(compile_tickers)
        return_db = return_db.set_index(['Ticker', 'Date']).unstack(0)

    # Unknown ticker request
    else:
        raise errors.MoabRequestError("Invalid window type")

    return (return_db)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT UNLESS YOUR NAME IS JACKSON!
# source: moabdb.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name='moabdb.proto',
    package='',
    syntax='proto3',
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x0cmoabdb.proto\"n\n\x07Request\x12\x0e\n\x06symbol\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61tatype\x18\x02 \x01(\t\x12\r\n\x05start\x18\x03 \x01(\r\x12\x0b\n\x03\x65nd\x18\x04 \x01(\r\x12\x10\n\x08username\x18\x05 \x01(\t\x12\r\n\x05token\x18\x06 \x01(\tJ\x04\x08\x07\x10\x10\"=\n\x08Response\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0cJ\x04\x08\x04\x10\x10\x62\x06proto3'
)


_REQUEST = _descriptor.Descriptor(
    name='Request',
    full_name='Request',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='symbol', full_name='Request.symbol', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='datatype', full_name='Request.datatype', index=1,
            number=2, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='start', full_name='Request.start', index=2,
            number=3, type=13, cpp_type=3, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='end', full_name='Request.end', index=3,
            number=4, type=13, cpp_type=3, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='username', full_name='Request.username', index=4,
            number=5, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='token', full_name='Request.token', index=5,
            number=6, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=16,
    serialized_end=126,
)


_RESPONSE = _descriptor.Descriptor(
    name='Response',
    full_name='Response',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='code', full_name='Response.code', index=0,
            number=1, type=13, cpp_type=3, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='message', full_name='Response.message', index=1,
            number=2, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='data', full_name='Response.data', index=2,
            number=3, type=12, cpp_type=9, label=1,
            has_default_value=False, default_value=b"",
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=128,
    serialized_end=189,
)

DESCRIPTOR.message_types_by_name['Request'] = _REQUEST
DESCRIPTOR.message_types_by_name['Response'] = _RESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Request = _reflection.GeneratedProtocolMessageType('Request', (_message.Message,), {
    'DESCRIPTOR': _REQUEST,
    '__module__': 'moabdb_pb2'
    # @@protoc_insertion_point(class_scope:Request)
})
_sym_db.RegisterMessage(Request)

Response = _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR': _RESPONSE,
    '__module__': 'moabdb_pb2'
    # @@protoc_insertion_point(class_scope:Response)
})
_sym_db.RegisterMessage(Response)


# @@protoc_insertion_point(module_scope)



import re
import requests
import pandas as pd
from datetime import datetime
from . import errors


def _to_unix_epoch(date_string):
    # Get uniq epoch time as integer
    unix_epoch = datetime.strptime(date_string, "%Y-%m-%d")
    tm = (unix_epoch - datetime(1970, 1, 1)).total_seconds()
    return int(tm)


def _to_unix_w_freq(sample_len, base_time, base_type):
    """
        Provide time frequency and base time, 
            returns offset time.
        Enter negative end_tm as base_time to find start date.
        Enter positive beg_tm as base_time to find end date.
    """
    tm_freq, tm_unit = re.findall(r'(\d+)(\w+?)', sample_len)[0]
    tm_freq = int(tm_freq) * (-1) if base_type == 'End' else int(tm_freq)

    base_timestamp = pd.Timestamp(base_time, unit='s')

    if str.upper(tm_unit) == 'D':
        new_time = base_timestamp + pd.DateOffset(days=tm_freq)
    elif str.upper(tm_unit) == 'W':
        new_time = base_timestamp + pd.DateOffset(weeks=tm_freq)
    elif str.upper(tm_unit) == 'M':
        new_time = base_timestamp + pd.DateOffset(months=tm_freq)
    elif str.upper(tm_unit) == 'Y':
        new_time = base_timestamp + pd.DateOffset(years=tm_freq)
    else:
        raise errors.MoabRequestError("Unknown time unit, accepts: D, W, M, Y")
    return (int(new_time.timestamp()))


def _get_unix_dates(sample_len, start_dt, end_dt):
    # User provided sample length ...

    # ... but didn't provide anything else --> get recent sample
    if (end_dt is None) & (start_dt is None):
        today = datetime.now().strftime("%Y-%m-%d")
        end = _to_unix_epoch(today)
        start = _to_unix_w_freq(sample_len, end, 'End')

    # ... and provided start date --> find end date using sample length
    elif (end_dt is None) & (start_dt is not None):
        start = _to_unix_epoch(start_dt)
        end = _to_unix_w_freq(sample_len, start, 'Start')

    # ... and provided end date --> find start date using sample length
    elif (end_dt is not None) & (start_dt is None):
        end = _to_unix_epoch(end_dt)
        start = _to_unix_w_freq(sample_len, end, 'End')

    # ... and provided start and end date --> ignore sample length
    elif (end_dt is not None) & (start_dt is not None):
        end = _to_unix_epoch(end_dt)
        start = _to_unix_epoch(start_dt)

    return (start, end)
